import nose

from django_anger.check_migration import check_duplicate_fields
from django_anger.check_migration import check_duplicate_models
from django_anger.check_migration import check_gratuitous_frozen_models
from django_anger.check_migration import check_missing_related_models
from django_anger.check_migration import check_model_names
from django_anger.check_migration import ValidationError


def check_model_names_test():
    check_model_names(open("testdata/good_migration.py"))
    with nose.tools.assert_raises(ValidationError):
        check_model_names(open("testdata/bad_migration_bad_model_name_1.py"))
    with nose.tools.assert_raises(ValidationError):
        check_model_names(open("testdata/bad_migration_bad_model_name_2.py"))


def check_duplicate_models_test():
    check_duplicate_models(open("testdata/good_migration.py"))
    with nose.tools.assert_raises(ValidationError):
        check_duplicate_models(open("testdata/bad_migration_model_frozen_twice.py"))


def check_duplicate_fields_test():
    check_duplicate_fields(open("testdata/good_migration.py"))
    with nose.tools.assert_raises(ValidationError):
        check_duplicate_fields(open("testdata/bad_migration_field_frozen_twice.py"))


def check_missing_foreign_keys_test():
    check_missing_related_models(open("testdata/good_migration.py"))
    with nose.tools.assert_raises(ValidationError):
        check_missing_related_models(open("testdata/bad_migration_missing_foreign_key.py"))


def check_gratuitous_frozen_models_test():
    check_gratuitous_frozen_models(open("testdata/good_migration.py"))
    with nose.tools.assert_raises(ValidationError):
        check_gratuitous_frozen_models(open("testdata/bad_migration_gratuitous_frozen_model.py"))



#!/usr/bin/env python

"""
Squash the initial migrations for multiple apps into a single combined
migration, avoiding potential issues with circular dependencies. See
ResettingMigrations.md for details.

Usage:
    # All apps in your Django project must have only an initial migration.
    # Say you want to save the squashed migration in app_alpha.
    squash_migrations app_alpha
"""

# TODO
# - validate migrations individually before squashing
# - if the same model is frozen in multiple apps, assert it's equal everywhere

from StringIO import StringIO
import os
import shutil
import textwrap

from django_anger.migration_utils import forwards_contents
from django_anger.migration_utils import parse_migration


def get_south_apps(project_dir):
    """
    Return a list of South-enabled apps in the given project_dir.

    Assumes an app is South-enabled if it has a 'migrations' directory.
    """
    apps = []
    for app in os.listdir(project_dir):
        dirname = os.path.join(project_dir, app)
        if not os.path.isdir(dirname):
            continue
        if os.path.exists(os.path.join(dirname, 'migrations')):
            apps.append(app)
    return apps


def get_migration_filenames(app_dir):
    """
    Return list of filenames of all migrations in the given app dir.
    """
    filenames = []
    migrations_dir = os.path.join(app_dir, 'migrations')
    for filename in os.listdir(migrations_dir):
        if not filename.endswith('.py'):
            continue
        if filename == '__init__.py':
            continue
        filenames.append(filename)
    return filenames


def get_path_of_sole_initial_migration(app_dir):
    """
    Return path to the sole 0001_initial migration in the given app dir.

    Raise ValueError if the app has other migrations.
    """
    filenames = get_migration_filenames(app_dir)
    if len(filenames) != 1 or filenames[0] != '0001_initial.py':
        raise ValueError("App '{}' has non-initial migrations."
                            .format(app_dir))
    return os.path.join(app_dir, 'migrations', filenames[0])


_imports = \
"""
# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models
"""

_class_def = \
"""
class Migration(SchemaMigration):
"""

_depends_on = \
"""
    depends_on = (
        ('{}', '{}'),
    )
"""

_forwards = \
"""
    def forwards(self, orm):
"""

_backwards = \
"""
    def backwards(self, orm):
        raise RuntimeError("Cannot reverse this migration.")
"""


def _pretty_print_models(models, output):
    """
    Custom pretty printer for models = { ... } part of a migration.
    """
    # Sadly pprint.pprint() looks ugly for this use case.
    print >>output, "\n    models = {"
    for model_name, model_def in sorted(models.items()):
        print >>output, "        '{}': {{".format(model_name)
        for field_name, field_def in sorted(model_def.items()):
            print >>output, "            '{}': {},".format(
                                field_name, field_def)
        print >>output, "        },"
    print >>output, "    }"


def _pretty_print_complete_apps(apps, output):
    """
    Custom pretty printer for complete_apps = [ ... ] part of a migration.
    """
    print >>output, "\n    complete_apps = ["
    for app in sorted(apps):
        print >>output, "        '{}',".format(app)
    print >>output, "    ]"


def squash_migrations(app_to_migration_path):
    """
    Squash many initial migrations into a single migration.

    `app_to_migration_path` is a dict from app name to the path of the initial
    migration for that app.

    Returns the squashed migration as a StringIO object.

    The squashed migration has:
    - no dependencies
    - a forwards() function obtained by concatenating the forwards() functions
      of all the given migrations
    - a backwards() function that raises RuntimeError
    - all frozen models in all the given migrations
    - all complete_apps in all the given migrations
    """
    output = StringIO()

    # Write header and start the forwards() method.
    print >>output, _imports
    print >>output, _class_def
    print >>output, _forwards

    # In forwards(), concatenate all the forwards() of individual migrations.
    for app, migration_path in sorted(app_to_migration_path.items()):
        print >>output, "\n        ### {} app ###\n".format(app)
        output.write(forwards_contents(open(migration_path)))

    # Write the backwards() method.
    print >>output, _backwards

    # Collect all frozen models from individual migrations, and write all of
    # them to the new migration.
    all_models = {}
    for app, migration_path in app_to_migration_path.iteritems():
        models, complete_apps = parse_migration(open(migration_path))
        if complete_apps != [app]:
            raise ValueError("App '{}' has unexpected complete_apps {}.".format(
                app, complete_apps))
        all_models.update(models)
    _pretty_print_models(all_models, output)

    # Write complete_apps for the new migration.
    _pretty_print_complete_apps(app_to_migration_path.keys(), output)

    return output


def make_dummy_migration(app, migration_path, destination_app,
                         squashed_migration_name):
    """
    Make a new initial migration for an app whose initial migration has been
    squashed.

    `app` is the name of the app for which to create an initial migration.
    `migration_path` is the path to the initial migration for that app.
    `destination_app` is the app containing the new squashed migration.
    `squashed_migration_name` is the name of the squashed migration (without
    the '.py' suffix)

    Returns the new initial migration as a StringIO object.

    The new migration has:
    - a dependency on the squashed migration
    - a forwards() function that does nothing
    - a backwards() function that raises RuntimeError
    - all frozen models in the original migration
    - the complete_apps in the original migration
    """
    output = StringIO()

    # Write header, depends_on, dummy forwards, and backwards.
    print >>output, _imports
    print >>output, _class_def
    print >>output, _depends_on.format(destination_app, squashed_migration_name)
    print >>output, _forwards
    print >>output, "        pass"
    print >>output, _backwards

    # Write the original frozen models and complete_apps for this migration.
    models, complete_apps = parse_migration(open(migration_path))
    _pretty_print_models(models, output)
    _pretty_print_complete_apps(complete_apps, output)

    return output


def main():
    import sys
    if len(sys.argv) < 2:
        print __doc__
        sys.exit(1)

    # Get all South-enabled apps.
    project_dir = os.getcwd()
    apps = get_south_apps(project_dir)
    print "Found the following apps:"
    for app in sorted(apps):
        print "    ", app
    print

    # Verify that all apps have only an initial migration.
    # Create dict from app to path of initial migration.
    app_to_migration_path = {}
    for app in apps:
        app_dir = os.path.join(project_dir, app)
        app_to_migration_path[app] = \
            get_path_of_sole_initial_migration(app_dir)

    # Verify that the given destination app is valid.
    destination_app = sys.argv[1]
    if destination_app not in apps:
        raise ValueError("Could not find app '{}' in current "
                         "directory.".format(destination_app))

    # Create the squashed migration and write it to disk.
    print "Squashing migrations..."
    squashed_migration_name = '0001_everything'
    squashed_migration_path = os.path.join(
        destination_app, 'migrations', squashed_migration_name + '.py')
    squashed_migration = squash_migrations(app_to_migration_path)
    with open(squashed_migration_path, 'w') as f:
        f.write(squashed_migration.getvalue())

    # Update the initial migration for every app.
    print "Updating initial migrations..."
    for app in apps:
        migration_path = app_to_migration_path[app]
        if app == destination_app:
            new_path = migration_path.replace('0001', '0002')
            shutil.move(migration_path, new_path)
            migration_path = new_path
        dummy_migration = make_dummy_migration(app, migration_path,
                                               destination_app,
                                               squashed_migration_name)
        with open(migration_path, 'w') as f:
            f.write(dummy_migration.getvalue())

    print "Done."
    print "The squashed migration is '{}'.".format(squashed_migration_path)



import ast
import re


def forwards_contents(f):
    """
    Return a string containing everything in the forwards() method of a
    migration.

    This string may not be safe to evaluate.
    """
    forwards_found = False
    forwards_lines = []
    for line in f:
        if 'def backwards(self, orm):' in line:
            break
        if forwards_found:
            forwards_lines.append(line)
        elif 'def forwards(self, orm):' in line:
            forwards_found = True
    return ''.join(forwards_lines)


def parse_migration(f):
    """
    Parse migration file and return (frozen_models, complete_apps).

    The returned objects are full-fledged Python data structures, as if we
    actually imported the migration file. But because we use ast.literal_eval(),
    this is safe to run on untrusted code.
    """
    # Collect lines defining models and complete_apps.
    # Example:
    # models = {
    #     ...
    # }
    # complete_apps = ['app_beta', 'app_gamma']
    model_lines = []
    complete_apps_lines = []
    models_found = False
    complete_apps_found = False
    for line in f:
        if not models_found and 'models =' in line:
            models_found = True
        if not complete_apps_found and 'complete_apps' in line:
            complete_apps_found = True
        if models_found and not complete_apps_found:
            model_lines.append(line)
        elif complete_apps_found:
            complete_apps_lines.append(line)
    if not models_found or not complete_apps_found:
        raise ValueError('Could not parse migration.')

    # Parse models.
    model_str = "".join(model_lines).strip()
    assert model_str.startswith('models =')
    model_str = model_str[len('models ='):]
    models = ast.literal_eval(model_str.strip())

    # Parse complete_apps.
    complete_apps_str = "".join(complete_apps_lines).strip()
    assert complete_apps_str.startswith('complete_apps =')
    complete_apps_str = complete_apps_str[len('complete_apps ='):]
    complete_apps = ast.literal_eval(complete_apps_str.strip())

    return models, complete_apps


def is_related(field):
    """
    Return True if the given field definition is a ForeignKey, OneToOneField, or
    ManyToManyField.

    >>> is_related(('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '100'}))
    False
    >>> is_related(('django.db.models.fields.related.ForeignKey', [], {'related_name': "'asdfadfa'", 'to': "orm['app_zeta.Model15']"}))
    True
    >>> is_related(('django.db.models.fields.related.OneToOneField', [], {'to': "orm['app_delta.Model09']", 'unique': 'True', 'null': 'True'}))
    True
    >>> is_related(('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['app_delta.Model13']", 'symmetrical': 'False'}))
    True
    """
    return field[0] in (
        'django.db.models.fields.related.ForeignKey',
        'django.db.models.fields.related.OneToOneField',
        'django.db.models.fields.related.ManyToManyField',
    )


def related_target(field):
    """
    Return target model of ForeignKey, OneToOneField, or ManyToManyField, as an
    "app.ModelName" string.

    >>> related_target(('django.db.models.fields.related.ForeignKey', [], {'related_name': "'asdfadfa'", 'to': "orm['app_zeta.Model15']"}))
    'app_zeta.Model15'
    >>> related_target(('django.db.models.fields.related.OneToOneField', [], {'to': "orm['app_delta.Model09']", 'unique': 'True', 'null': 'True'}))
    'app_delta.Model09'
    >>> related_target(('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['app_delta.Model13']", 'symmetrical': 'False'}))
    'app_delta.Model13'
    """
    to_re = re.compile("^orm\['([^']+)'\]$")
    return to_re.match(field[2]['to']).group(1)




#!/usr/bin/env python

"""
Parse a South migration file, and do some basic sanity checking on it.

Usage:
    check_migration my_migration.py
"""

from django_anger.migration_utils import is_related
from django_anger.migration_utils import parse_migration
from django_anger.migration_utils import related_target


class ValidationError(Exception):
    pass


validators = []


def register_validator(func):
    validators.append(func)
    return func


@register_validator
def check_missing_related_models(f):
    """
    Verify that for all related models, the target model is also frozen.
    """
    models, complete_apps = parse_migration(f)
    for model, fields in models.iteritems():
        for field_name, field_def in fields.iteritems():
            if field_name == 'Meta':
                continue
            if not is_related(field_def):
                continue
            target = related_target(field_def).lower()
            if target not in models:
                raise ValidationError(
                    "Field '{}' has ForeignKey to model '{}', which is "
                    "not frozen.".format(field_name, target))


@register_validator
def check_duplicate_models(f):
    """
    Verify that no model is frozen twice.
    """
    # parse_migration() gives us a dict, which cannot have duplicate keys.
    # So we have to get messy and look at the raw file.
    models, complete_apps = parse_migration(f)
    found = {}
    f.seek(0)
    for line in f:
        for model in models.keys():
            # Fragile string-based matching.
            if "'{}':".format(model) in line:
                if model in found:
                    raise ValidationError(
                            "Model '{}' frozen twice.".format(model))
                found[model] = True


@register_validator
def check_duplicate_fields(f):
    """
    Verify that no model has the same field frozen twice.
    """
    # parse_migration() gives us dicts, which cannot have duplicate keys.
    # So we have to get messy and look at the raw file.
    models, complete_apps = parse_migration(f)
    current_model = None
    fields_found = {}
    f.seek(0)
    for line in f:
        # Check if this line starts a new model definition.
        for model in models.keys():
            # Fragile string-based matching.
            if "'{}':".format(model) in line:
                current_model = model
                fields_found = {}
        # Check if this line is a field definition.
        if current_model:
            for field in models[current_model].keys():
                # Fragile string-based matching.
                if "'{}':".format(field) in line:
                    if field in fields_found:
                        raise ValidationError(
                            "Field '{}' of model '{}' frozen twice.".format(
                                field, current_model))
                    fields_found[field] = True


@register_validator
def check_model_names(f):
    """
    Verify that the frozen model name is lowercase and matches the capitalized
    model name in Meta.
    """
    models, complete_apps = parse_migration(f)
    for model, fields in models.iteritems():
        app_name, lowercase_model_name = model.split('.')
        capitalized_model_name = fields['Meta']['object_name']
        if lowercase_model_name.lower() != lowercase_model_name:
            raise ValidationError("Expected lower-case app.model_name, but "
                                  "found '{}'.".format(lowercase_model_name))
        if capitalized_model_name.lower() != lowercase_model_name:
            raise ValidationError("Meta object_name '{}' does not match "
                                  "expected model name '{}'.".format(
                                      capitalized_model_name,
                                      lowercase_model_name))


@register_validator
def check_gratuitous_frozen_models(f):
    """
    Verify that a model is frozen only if it is in a complete_apps app, or if it
    is referenced by a ForeignKey.
    """
    models, complete_apps = parse_migration(f)
    related_targets = set()
    for model, fields in models.iteritems():
        for field_name, field_def in fields.iteritems():
            if field_name == 'Meta':
                continue
            if is_related(field_def):
                related_targets.add(related_target(field_def).lower())
    for model in models:
        app, model_name = model.split('.')
        if app not in complete_apps and model not in related_targets:
            raise ValidationError(
                    "Model '{}' frozen but not used.".format(model))


def validate_migration_file(f):
    """
    Run all the validators in sequence, propagating any exceptions they might
    raise.
    """
    for validator in validators:
        validator(f)
        f.seek(0)


def main():
    import sys
    if len(sys.argv) < 2:
        print __doc__
        sys.exit(1)

    for migration in sys.argv[1:]:
        print "Validating {}...".format(migration)
        try:
            f = open(migration)
            validate_migration_file(f)
            print " -> OK."
        except Exception as err:
            print " -> failed: {}".format(err)
            print "\nFull traceback:\n"
            import traceback
            traceback.print_exc()
            sys.exit(1)

    print "\nAll is well in the world."



import distribute_setup
distribute_setup.use_setuptools()

from setuptools import setup

from django_anger import __version__


setup(
    name = "django-anger",
    version = __version__,
    packages = ["django_anger"],
    py_modules = ["distribute_setup"],

    # metadata for upload to PyPI
    author = "Constantin Berzan",
    author_email = "cberzan@gmail.com",
    description = "tools and hacks for using Django _in anger_",
    # long_description = 
    license = "MIT",
    keywords = "django south migration migrations",
    url = "https://github.com/cberzan/django-anger",
    # download_url =

    classifiers = [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
    ],

    entry_points = {
        'console_scripts': [
            'check_migration = django_anger.check_migration.main',
            'squash_migrations = django_anger.squash_migrations.main',
            'migration_strings = django_anger.migration_strings.main',
        ],
    },
)
